def __turn_two_demension_dict_into_one_demension_combination(
    two_demension_dict: dict
) -> dict:
    res = {}
    for node_1, _dict in two_demension_dict.items():
        for node_2, value in _dict.items():
            if ((node_1, node_2) in res) or ((node_2, node_1) in res):
                continue
            res[node_1, node_2] = value
    return res

def __turn_two_demension_dict_into_one_demension_permutation(
    two_demension_dict: dict
) -> dict:
    res = {}
    for node_1, _dict in two_demension_dict.items():
        for node_2, value in _dict.items():
            if ((node_1, node_2) in res):
                continue
            res[node_1, node_2] = value
    return res

def __delete_repeat_key(one_demension_dict: dict
) -> dict:
    key_to_delete = []
    for node1, node2 in one_demension_dict:
        if node1 == node2:
            key_to_delete.append((node1, node2))
    for key in key_to_delete:
        del one_demension_dict[key]

def turn_two_demension_dict_into_one_demension(
    two_demension_dict: dict, 
    method: str="permutation",
    repeat: bool=False
) -> dict:
    """
    将二维字典转成一维字典，例：
    {"x": {"y": 10}}  =>  {("x", "y"): 10}
    two_demension_dict - 二维字典
    method - 选择以“组合”或“排列”的方式合并
    repeat - 相同的顶点是否允许组合
    """
    if method == "permutation":
        res = __turn_two_demension_dict_into_one_demension_permutation(two_demension_dict)
    elif method == "combination":
        res = __turn_two_demension_dict_into_one_demension_combination(two_demension_dict)
    else:
        raise Exception(f"Unsupportable param \"method\" {method}.")
    if not repeat:
        __delete_repeat_key(res)
    return res

def __check_vector_length(*args: list
) -> bool:
    """
    检查多个列表的长度是否一致
    """
    length_list = []
    for vector in args:
        length_list.append(len(vector))
    return length_list[0] == (sum(length_list) // len(length_list))

def v_greater_than(vector_1: list, vector_2: list
) -> list:
    """
    比较向量1是否大于向量2，即比较向量1的对应元素是否大于向量2，
    如果大于则为True，否则为False
    将所有比较结果打包在一个列表中作为返回值
    vector_1 - 向量1 
    vector_2 - 向量2
    """
    if __check_vector_length(vector_1, vector_2):
        res_bool_list = []
        for element_1, element_2 in zip(vector_1, vector_2):
            if element_1 >= element_2:
                res_bool_list.append(True)
            else:
                res_bool_list.append(False)
        return res_bool_list
    else:
        raise Exception("Length is not equal between vector_1 and vector_2.")



import asyncio
from asyncio import AbstractEventLoop
from threading import Lock, Thread


class LoopManagerMeta(type):
    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class LoopManager(metaclass=LoopManagerMeta):
    _loops = {}  # {"ThreadName": loop object}

    def start_new_loop(self, thread_name):
        new_loop = asyncio.new_event_loop()

        def new_thread(new_loop: AbstractEventLoop):
            asyncio.set_event_loop(new_loop)
            new_loop.run_forever()

        new_thread = Thread(target=new_thread, args=(new_loop,), name=thread_name,)
        new_thread.setDaemon(True)
        new_thread.start()
        return new_thread, new_loop

    def new_event_loop(self, thread_name: str = None) -> AbstractEventLoop:
        new_thread, new_loop = self.start_new_loop(thread_name)
        self._loops.update({new_thread.getName(): new_loop})
        return new_loop

    def get_event_loop(self, thread_name=None):
        if thread_name is None:
            return self._loops
        else:
            return self._loops.get(thread_name)




from pydantic import BaseModel
from typing import Any, Dict, List, Optional
from enum import Enum
import functools


class NoteConfig(BaseModel):
    directory_name: Optional[str]
    filename_prefix: Optional[List[str]]
    header: Optional[List[str]]


class FileType(Enum):
    csv = "csv"
    excel = "excel"
    pickle = "pickle"


class Result(BaseModel):
    persistent: Optional[Dict[FileType, Dict[str, Any]]]
    result: Optional[Dict[FileType, Dict[str, Any]]]
    note: Optional[Dict[str, Any]]


class Note:
    filepath: Optional[str] = None
    header: Optional[List[str]] = None
    encoding: str = "utf-8"

    @classmethod
    def __check_file(cls):
        from os.path import exists

        if not exists(cls.filepath):
            return False
        return True

    @classmethod
    def __check_header(cls):
        header = ",".join(cls.header)
        with open(cls.filepath, "r", encoding=cls.encoding) as fr:
            same_header = header == fr.readline()
        if not same_header:
            return False
        return True

    @classmethod
    def set_config(cls, note_config: NoteConfig):
        from os.path import join
        from os import getcwd
        from datetime import datetime

        if note_config.filename_prefix and note_config.directory_name:
            file_name = "_".join(
                note_config.filename_prefix
                + [datetime.strftime(datetime.now(), "%Y%m%d%H%M%S") + ".csv",]
            )
            cls.filepath = join(getcwd(), note_config.directory_name, file_name)
            cls.header = note_config.header
            if (not cls.__check_file()) or (not cls.__check_header()):
                header = ",".join(cls.header) + "\n"
                cls.record_line(header)

    @classmethod
    def record_line(cls, line):
        with open(cls.filepath, "a+", encoding=cls.encoding) as fa:
            fa.write(line)

    @classmethod
    def record(cls, note_config: NoteConfig):
        cls.set_config(note_config)

        def wrapper(func):
            @functools.wraps(func)
            def inner(*args, **kwargs):
                result: Result = func(*args, **kwargs)
                line = []
                for column_name in cls.header:
                    line.append(str(result.note[column_name]))
                line = ",".join(line) + "\n"
                cls.record_line(line=line)

            @functools.wraps(func)
            def inner_empty(*args, **kwargs):
                return func(*args, **kwargs)

            if cls.filepath != None and cls.header != None:
                return inner
            else:
                return inner_empty

        return wrapper
